/*
 * @(#)ArrayList.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.compatibility;

import java.util.Vector;

/**
 * The Java 1.1 implementation of ArrayList.
 * @version 1.0 2004.01.11
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
class ArrayList_1_1 implements List {

  /** The vector. */
  protected Vector vector;

  /**
   * Constructs an empty list with an initial capacity of ten.
   */
  public ArrayList_1_1() {
    vector = new Vector();
  }

  /**
   * Constructs a list containing the elements of the specified collection, in
   * the order they are returned by the collection's iterator.
   * @param c The collection.
   */
  public ArrayList_1_1(Collection c) {
    vector = new Vector(c.size());
    Object[] objects = c.toArray();
    for(int i=0; i<objects.length; i++) {
      add(objects[i]);
    }
  }

  /**
   * Constructs an empty list with the specified initial capacity.
   * @param initialCapacity The initial capacity.
   */
  public ArrayList_1_1(int initialCapacity) {
    vector = new Vector(initialCapacity);
  }

  /**
   * Returns an array containing all of the elements in this collection. 
   * @return The array.
   */
  public Object[] toArray() {
    Object[] objects = new Object[vector.size()];
    vector.copyInto(objects);
    return objects;
  }

  /**
   * Returns an array containing all of the elements in this collection; the
   * runtime type of the returned array is that of the specified array.
   * @param a The array from which to get the type.
   * @return The array.
   */
  public Object[] toArray(Object[] a) {
    int elementCount = size();
    Object[] elementData = toArray();
    if (a.length < elementCount) {
      a = (Object[])java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), elementCount);
    }
    System.arraycopy(elementData, 0, a, 0, elementCount);
    if (a.length > elementCount) {
      a[elementCount] = null;
    }
    return a;
  }

  /**
   * Returns true if this collection contains the specified element.
   * @param o The object to look for.
   * @return True if found.
   */
  public boolean contains(Object o) {
    return vector.contains(o);
  }

  /**
   * Appends the specified element to the end of this list (optional operation).
   * @param o The object to add.
   * @return True.
   */
  public boolean add(Object o) {
    vector.addElement(o);
    return true;
  }

  /**
   * Inserts the specified element at the specified position in this list
   * (optional operation).
   * @param index The index.
   * @param element The element.
   */
  public void add(int index, Object element) {
    vector.insertElementAt(element, index);
  }

  /**
   * Removes the element at the specified position in this list (optional
   * operation).
   * @param index The index of the element to remove.
   * @return The element that was removed.
   */
  public Object remove(int index) {
    Object element = vector.elementAt(index);
    vector.removeElementAt(index);
    return element;
  }

  /**
   * Removes the first occurrence in this list of the specified element
   * (optional operation). 
   * @param o The element to remove.
   * @return True if the element was present.
   */
  public boolean remove(Object o) {
    return vector.removeElement(o);
  }

  /**
   * Removes all this collection's elements that are also contained in the
   * specified collection (optional operation).
   * @param c The collection of elements to remove.
   * @return True if the collection was affected by this operation.
   */
  public boolean removeAll(Collection c) {
    Object[] objects = c.toArray();
    boolean isChanged = false;
    for(int i=0; i<objects.length; i++) {
      isChanged |= remove(objects[i]);
    }
    return isChanged;
  }

  /**
   * Replaces the element at the specified position in this list with the
   * specified element (optional operation).
   * @param index The index at which to set the element.
   * @param element The element to set.
   * @return The element previously located at the specified position.
   */
  public Object set(int index, Object element) {
    return vector.set(index, element);
  }

  /**
   * Returns the number of elements in this list.
   * @return The size.
   */
  public int size() {
    return vector.size();
  }

  /**
   * Returns the element at the specified position in this list.
   * @param index The index.
   * @return The element.
   */
  public Object get(int index) {
    return vector.elementAt(index);
  }

  /**
   * Returns the index in this list of the first occurrence of the specified
   * element, or -1 if this list does not contain this element.
   * @param o The element to look for.
   * @return The index.
   */
  public int indexOf(Object o) {
    return vector.indexOf(o);
  }

  /**
   * Returns true if this list contains no elements.
   * @return True if this list is empty.
   */
  public boolean isEmpty() {
    return vector.isEmpty();
  }

  /**
   * Adds all of the elements in the specified collection to this set if they're
   * not already present (optional operation).
   * @param c The collection.
   * @return True if changed.
   */
  public boolean addAll(Collection c) {
    Object[] objects = c.toArray();
    vector.ensureCapacity(objects.length);
    boolean isChanged = false;
    for(int i=0; i<objects.length; i++) {
      isChanged |= add(objects[i]);
    }
    return isChanged;
  }

  /**
   * Increases the capacity of this ArrayList instance, if necessary, to ensure
   * that it can hold at least the number of elements specified by the minimum
   * capacity argument.
   * @param minCapacity The minimum capacity.
   */
  public void ensureCapacity(int minCapacity) {
    vector.ensureCapacity(minCapacity);
  }

}
