/*
 * @(#)Accessor.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy;

import java.awt.Component;

import chrriis.uihierarchy.compatibility.HashSet;
import chrriis.uihierarchy.compatibility.List;
import chrriis.uihierarchy.compatibility.Set;

import chrriis.uihierarchy.accessor.AccessorRunnable;

/**
 * An accessor to access and manipulate the hierarchy.
 * @version 1.0 2003.12.02
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class Accessor {

  /** The configuration the accessor uses. */
  protected UIHConfig uihConfig;

  /** The levels that are manipulated by this accessor. */
  protected HLevel[] hLevels;

  /**
   * Construct an Accessor.
   * @param uihConfig The configuration that this accessor is using.
   * @param hLevels The levels to be manipulated by the accessor.
   */
  protected Accessor(UIHConfig uihConfig, HLevel[] hLevels) {
    this.uihConfig = uihConfig;
    this.hLevels = hLevels;
  }

  /**
   * Get the configuration the accessor uses.
   * @return The configuration.
   */
  protected UIHConfig getConfig() {
    return uihConfig;
  }

  /**
   * Get the levels manipulated by this accessor.
   * @return The levels.
   */
  protected HLevel[] getHLevels() {
    return hLevels;
  }

  /**
   * Get the components at these hierarchy levels.
   * @return The components.
   */
  public Component[] getComponents() {
    HLevel[] hLevels = getHLevels();
    Component[] components = new Component[hLevels.length];
    for(int i=0; i<components.length; i++) {
      components[i] = hLevels[i].getComponent();
    }
    return components;
  }

/*----------------------------------------------------------------------------*/

  /**
   * Aquire the next free slot.
   * @return The next free slot.
   */
  protected static int aquireSlot() {
    return HLevel.aquireSlot();
  }

  /**
   * Get the object at the particular slot of a level.
   * @param hLevel The level from which to get the slot object.
   * @param slot The slot from where to get the object.
   * @return The object that was stored at the indicated slot.
   */
  protected static Object getSlotObject(HLevel hLevel, int slot) {
    return hLevel.getSlotObject(slot);
  }

  /**
   * Set the object at a given slot of a level.
   * @param hLevel The level for which to set the slot object.
   * @param slotObject The object to set.
   * @param slot The slot at which to set the object.
   */
  protected static void setSlotObject(HLevel hLevel, Object slotObject, int slot) {
    hLevel.setSlotObject(slotObject, slot);
  }

/*----------------------------------------------------------------------------*/

  /** The slot of the disabling IDs. */
  protected static final int DISABLING_IDS_SLOT = aquireSlot();

  /**
   * Force the components to be enabled. This discards all the disabled states
   * set using some disabling IDs.
   */
  public void forceEnabled() {
    forceEnabled(false);
  }

  /**
   * Force the hierarchy of components to be enabled. This discards all the
   * disabled states set using some disabling IDs.
   */
  public void forceHEnabled() {
    forceEnabled(true);
  }

  /**
   * Force the components to be enabled. This discards all the disabled states
   * set using some disabling IDs.
   * @param isTraversing Indicate if it should also perform the same action on
   * the sub levels.
   */
  protected void forceEnabled(boolean isTraversing) {
    HLevel[] hLevels = getHLevels();
    for(int i=0; i<hLevels.length; i++) {
      HLevel hLevel = hLevels[i];
      if(isTraversing && hLevel instanceof HParentLevel) {
        List childrenList = ((HParentLevel)hLevel).getChildrenList();
        for(int j=0; j<childrenList.size(); j++) {
          getConfig().createAccessor((HLevel)childrenList.get(j)).forceEnabled(isTraversing);
        }
      }
      setSlotObject(hLevel, null, DISABLING_IDS_SLOT);
      hLevel.getComponent().setEnabled(true);
    }
  }

  /**
   * Set whether the components are enabled, using the default disabling ID of
   * 1. The components are enabled when all the groups that set the state set it
   * to enabled.
   * @param isEnabled True if the component is to be enabled.
   */
  public void setEnabled(boolean isEnabled) {
    setEnabled(isEnabled, 1);
  }

  /**
   * Set whether the components are enabled with regard to the disabling ID. The
   * components are enabled when all the groups that set the state set it to
   * enabled.
   * @param isEnabled True if the component is to be enabled.
   * @param id The disabling ID.
   */
  public void setEnabled(boolean isEnabled, int id) {
    setEnabled(isEnabled, id, false);
  }

  /**
   * Set whether the hierarchy is enabled. Enabling/disabling the hierarchy has
   * the same effect as calling the <code>setEnabled</code> method on the
   * sub-levels recursively, using the default disabling ID of -1.
   * @param isHEnabled True if the hierarchy is to be enabled.
   */
  public void setHEnabled(boolean isHEnabled) {
    setHEnabled(isHEnabled, -1);
  }

  /**
   * Set whether the hierarchy is enabled. Enabling/disabling the hierarchy has
   * the same effect as calling the <code>setEnabled</code> method on the
   * sub-levels recursively.
   * @param isHEnabled True if the hierarchy is to be enabled.
   * @param id The disabling ID.
   */
  public void setHEnabled(boolean isHEnabled, int id) {
    setEnabled(isHEnabled, id, true);
  }

  /**
   * Set whether the components are enabled with regard to the disabling ID. The
   * components are enabled when all the groups that set the state set it to
   * enabled.
   * @param isEnabled True if the component is to be enabled.
   * @param id The disabling ID.
   * @param isTraversing Indicate if it should also perform the action on the
   * sub-levels.
   */
  protected void setEnabled(boolean isEnabled, int id, boolean isTraversing) {
    HLevel[] hLevels = getHLevels();
    for(int i=0; i<hLevels.length; i++) {
      HLevel hLevel = hLevels[i];
      if(isTraversing && hLevel instanceof HParentLevel) {
        List childrenList = ((HParentLevel)hLevel).getChildrenList();
        for(int j=0; j<childrenList.size(); j++) {
          getConfig().createAccessor((HLevel)childrenList.get(j)).setEnabled(isEnabled, id, isTraversing);
        }
      }
      Set disabledIDsSet = (Set)getSlotObject(hLevel, DISABLING_IDS_SLOT);
      if(isEnabled) {
        if(disabledIDsSet != null) {
          disabledIDsSet.remove(new Integer(id));
          if(disabledIDsSet.isEmpty()) {
            disabledIDsSet = null;
            setSlotObject(hLevel, null, DISABLING_IDS_SLOT);
          }
        }
        if(disabledIDsSet == null) {
          hLevel.getComponent().setEnabled(isEnabled);
        }
      } else {
        if(disabledIDsSet == null) {
          disabledIDsSet = new HashSet(1);
          setSlotObject(hLevel, disabledIDsSet, DISABLING_IDS_SLOT);
        }
        disabledIDsSet.add(new Integer(id));
        hLevel.getComponent().setEnabled(isEnabled);
      }
    }
  }

  /**
   * Indicate if the components are maybe enabled, using -1 as the
   * default disabling ID. If the accessor manipulates more than one level, this
   * is just an indication as some levels may be configured using the same IDs
   * from other accessors.
   * @return The state of the components.
   */
  public boolean isEnabled() {
    return isEnabled(1);
  }

  /**
   * Indicate if the hierarchy of components are maybe enabled, using -1 as the
   * default disabling ID. If the accessor manipulates more than one level, this
   * is just an indication as some levels may be configured using the same IDs
   * from other accessors.
   * @return The state of the hierarchy of components.
   */
  public boolean isHEnabled() {
    return isEnabled(-1);
  }

  /**
   * Indicate if the components are maybe enabled, considering all the disabling
   * IDs in effect. If the accessor manipulates more than one level, this
   * is just an indication as some levels may be configured from other
   * accessors.
   * @return The state of the components.
   */
  public boolean getEnabledState() {
    return getHLevels()[0].getComponent().isEnabled();
  }

  /**
   * Indicate if the components or hierarchy of components are/is maybe enabled
   * with regard to a specific ID. If the accessor manipulates more than one
   * level, this is just an indication as some levels may be configured using
   * the same IDs from other accessors.
   * @param id The disabling ID.
   * @return The state of the components or hierarchy of components.
   */
  public boolean isEnabled(int id) {
    Set disabledIDsSet = (Set)getSlotObject(getHLevels()[0], DISABLING_IDS_SLOT);
    return(!(disabledIDsSet != null && disabledIDsSet.contains(new Integer(id))));
  }

/*----------------------------------------------------------------------------*/

  /**
   * Set the visibility of the manipulated levels.
   * @param isVisible True if the levels should be shown.
   */
  public void setVisible(final boolean isVisible) {
    run(new AccessorRunnable() {
      public Object run(Component component, Accessor childrenAccessor) {
        component.setVisible(isVisible);
        return null;
      }
    });
  }

  /**
   * Indicate if the components of the levels may be visible. If the accessor
   * manipulates more than one level, this is just an indication as the levels
   * may be configured inconsistently.
   * @return The visibility of the components.
   */
  public boolean isVisible() {
    return getHLevels()[0].getComponent().isVisible();
  }

/*----------------------------------------------------------------------------*/

  /**
   * Run an action on all the components of the levels of this accessor.
   * @param accessorRunnable The action to run.
   * @param isTraversing Indicate whether children of each levels should be
   * traversed. This traversal occurs before running for the parent.
   * @return An array of the results generated by the calls on the levels of
   * this accessor, but not the sub levels (if using traversal).
   */
  protected Object[] run(AccessorRunnable accessorRunnable, boolean isTraversing) {
    HLevel[] hLevels = getHLevels();
    Object[] results = new Object[hLevels.length];
    for(int i=0; i<hLevels.length; i++) {
      HLevel hLevel = hLevels[i];
      List childrenList = null;
      if(hLevel instanceof HParentLevel) {
        if(isTraversing) {
          childrenList = ((HParentLevel)hLevel).getChildrenList();
          for(int j=0; j<childrenList.size(); j++) {
            getConfig().createAccessor((HLevel)childrenList.get(j)).run(accessorRunnable, isTraversing);
          }
        }
      }
      Accessor childrenAccessor = null;
      if(childrenList != null && !childrenList.isEmpty()) {
        HLevel[] childrenHLevels = new HLevel[childrenList.size()];
        for(int j=0; j<childrenHLevels.length; j++) {
          childrenHLevels[j] = (HLevel)childrenList.get(j);
        }
        childrenAccessor = getConfig().createAccessor(childrenHLevels);
      }
      results[i] = accessorRunnable.run(hLevel.getComponent(), childrenAccessor);
    }
    return results;
  }

  /**
   * Run an action on all the components of the levels of this accessor, and
   * their sub hierarchies.
   * @param accessorRunnable The action to run.
   * @return An array of the results generated by the calls on the levels of
   * this accessor, but not the sub levels.
   */
  public Object[] runTraversing(AccessorRunnable accessorRunnable) {
    return run(accessorRunnable, true);
  }

  /**
   * Run an action on all the components of the levels of this accessor.
   * @param accessorRunnable The action to run.
   * @return An array of the results generated by the calls on the levels of
   * this accessor.
   */
  public Object[] run(AccessorRunnable accessorRunnable) {
    return run(accessorRunnable, false);
  }

/*----------------------------------------------------------------------------*/

  /**
   * Remove the manipulated levels from their hierarchy. This operation clears
   * the constraints that were set for the levels, so they will have to be
   * redeclared when adding to a new hierarchy.
   */
  public void remove() {
    HLevel[] hLevels = getHLevels();
    for(int i=0; i<hLevels.length; i++) {
      HLevel hLevel = hLevels[i];
      hLevel.remove();
      hLevel.clearConstraints();
    }
  }

}
