/*
 * @(#)Unit.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.unit;

import chrriis.uihierarchy.compatibility.HashMap;
import chrriis.uihierarchy.compatibility.Map;

/**
 * A unit, used to perform conversions, mainly from and to pixels.
 * @version 1.0 2005.02.01
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public abstract class Unit {

  /**
   * Construct a unit.
   */
  protected Unit() {
  }

  /**
   * Get the ratio to use when making conversions. The ratio is so that:<br>
   * <code>unitValue * ratio == pixelValue</code>.
   * @return The ratio to use for conversions.
   */
  protected abstract double getRatio();

  /**
   * Convert a value in a given unit to the value in this unit.
   * @param unit The unit of the value to convert.
   * @param value The value to convert.
   * @return The result of the conversion of the value.
   */
  protected double convert(Unit unit, double value) {
    if(unit == this) {
      return value;
    }
    return unit.getRatio() * value / getRatio();
  }

  /**
   * Get the names representing the unit.
   * @return An array containing the names.
   */
  public abstract String[] getNames();

  /** The map to associate some names to some units. */
  protected static Map nameToUnitMap;

  /**
   * Register a unit, using the names it defines.
   * @param unit The unit to register.
   */
  public static void registerUnit(Unit unit) {
    registerUnit(unit, unit.getNames());
  }

  /**
   * Register a unit, with some names that define it.
   * @param unit The unit to register.
   * @param names The names that identify the unit.
   */
  public static void registerUnit(Unit unit, String[] names) {
    if(names == null || names.length == 0) {
      throw new IllegalStateException("The unit must have some identifying names!");
    }
    for(int i=0; i<names.length; i++) {
      String name = names[i];
      if(name == null || "".equals(name)) {
        throw new IllegalStateException("An identifying name of a unit cannot be null or empty!");
      }
      if(nameToUnitMap == null) {
        nameToUnitMap = new HashMap();
      }
      Unit oldUnit = (Unit)nameToUnitMap.get(name);
      if(oldUnit != null && !oldUnit.equals(unit)) {
        System.err.println("A unit with the name \"" + name + "\" is already defined! Skipping addition.");
      } else {
        nameToUnitMap.put(name, unit);
      }
    }
  }

  /**
   * Get a unit registered with the given name.
   * @param unitName A name identifying the unit to get.
   * @return The unit corresponding to that name, or null if could not be found.
   */
  public static Unit getUnit(String unitName) {
    if(unitName == null) {
      throw new IllegalArgumentException("The name of the unit to retrieve cannot be null!");
    }
    if(nameToUnitMap == null) {
      return null;
    }
    return (Unit)nameToUnitMap.get(unitName);
  }

  static {
    registerUnit(Pixel.getUnit());
    registerUnit(DluX.getUnit());
    registerUnit(DluY.getUnit());
    registerUnit(DtpPoint.getUnit());
    registerUnit(Inch.getUnit());
    registerUnit(Millimeter.getUnit());
    registerUnit(Centimeter.getUnit());
  }

  /**
   * Parse a value along with its unit from a String representation, and
   * convert it to its value in pixels.
   * @param unitValue The value with a unit in its String
   * representation. If the unit is not specified, then assume it is Pixel.
   * @return The double value resulting from the conversion.
   */
  protected static double parse(String unitValue) {
    if(unitValue == null) {
      throw new IllegalArgumentException("The representation of a unit cannot be null!");
    }
    unitValue = unitValue.trim();
    int index = unitValue.length();
    for(; index > 0 && Character.isLetter(unitValue.charAt(index - 1)); index--);
    String value = unitValue.substring(0, index).trim();
    String unit = unitValue.substring(index).trim();
    if("".equals(unit)) {
      return Double.parseDouble(value);
    }
    Unit sourceUnit = Unit.getUnit(unit);
    if(sourceUnit == null) {
      throw new IllegalArgumentException("The unit \""+ unit + "\" is not registered!");
    }
    return Pixel.doubleValue(sourceUnit, Double.parseDouble(value));
  }

}
