/*
 * @(#)XmlUIH.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.xml;

import java.awt.LayoutManager;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Enumeration;
import java.util.Hashtable;

import nanoxml.XMLElement;

import chrriis.uihierarchy.HLeaf;
import chrriis.uihierarchy.HNode;
import chrriis.uihierarchy.HRootNode;
import chrriis.uihierarchy.UIH;
import chrriis.uihierarchy.compatibility.ArrayList;
import chrriis.uihierarchy.compatibility.List;
import chrriis.uihierarchy.xml.resolver.ArrayComponentResolver;
import chrriis.uihierarchy.xml.resolver.ComponentResolver;


/**
 * A builder of UI hierarchy from XML documents. It parses XML documents and
 * creates their hierarchies using the <code>UIH</code> class.
 * <br><br>
 * The DTD file "uih.dtd" indicates what are the possible tags. The basic idea
 * is that where you would call the <code>openNew</code> method there is a tag
 * <code>&lt;rootNode&gt;</code> with attributes or children entities
 * representing the possible method calls of the node that is added. It is the
 * same for the <code>open</code> and <code>add</code> methods where the
 * corresponding tags are <code>&lt;node&gt;</code> and
 * <code>&lt;leaf&gt;</code>.
 * <br><br>
 * One important point that is worth to detail is the way parameters work: by
 * default they are supposed to be Strings. Let's look at this example:
 * <br><br>
 * <code>
 * &lt;node value="A label: " constraints="gridx=0, gridy=0"/&gt;
 * </code>
 * <br><br>
 * The constraints are a String. This next example is equivalent:
 * <br><br>
 * <code>
 * &lt;node value="A label: "&gt;<br>
 * &nbsp;&nbsp;&lt;constraints class="String" value="gridx=0, gridy=0"&gt;<br>
 * &lt;/node&gt;
 * </code>
 * <br><br>
 * From the above example, we can see that any type of data can be used,
 * provided they have a constructor using a String, as the <code>value</code>
 * will be given to the constructor as a String.
 * <br><br>
 * We can also notice that "java.lang" classes or base types do not need the
 * full package name. Also the package name can be omitted for known layouts or
 * for layouts with a name ending with "Layout" and located in the "java.awt"
 * package.
 * <br>
 * Constraints are using this mechanism, as well as subConstraints and
 * layoutParam.
 * <br><br>
 * An addition to layoutParam is the ability to provide more than one attribute.
 * Let's consider an example where the layout to use is a CardLayout and we need
 * to construct it with two parameters of type <code>int</code>. The result
 * would be:
 * <br><br>
 * <code>
 * &lt;node layout="CardLayout"&gt;<br>
 * &nbsp;&nbsp;&lt;layoutParam&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;param class="int" value="20"/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;param class="int" value="20"/&gt;<br>
 * &nbsp;&nbsp;&lt;/layoutParam&gt;<br>
 * ...
 * </code>
 * <br><br>
 * Internally this code generates an array with two elements. If you need an
 * array of arrays, just declare them using some children <code>
 * &lt;param&gt;</code> tags:
 * <br><br>
 * <code>
 * ...<br>
 * &nbsp;&nbsp;&lt;param&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;param class="String" value="s1"/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;param class="Integer" value="i1"/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;param&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&lt;param class="String" value="s2"/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;/param&gt;<br>
 * &nbsp;&nbsp;&lt;/param&gt;<br>
 * ...
 * </code>
 * <br><br>
 * Note that this is not extensively tested for now, and only supports Object[]
 * arrays (no support for String[] or other types).
 * @version 1.0 2003.12.08
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class XmlUIH {

  /** The component resolver. */
  protected ComponentResolver componentResolver;

  /**
   * Get the component resolver that was used when creating the hierarchy, to
   * map the names in the XML document to some real components.
   * @return The component resolver. It can be null if the hierarchy is not
   * yet created.
   */
  public ComponentResolver getComponentResolver() {
    return componentResolver;
  }

  /**
   * Set the component resolver to use when some components will need to be resolved from names.
   * @param componentResolver The component resolver.
   */
  protected void setComponentResolver(ComponentResolver componentResolver) {
    this.componentResolver = componentResolver;
  }

  /** The name of the hierarchy to load. */
  protected String hierarchyName;

  /**
   * Get the name of the hierarchy to use.
   * @return The name of the hierarchy to use.
   */
  public String getHierarchyName() {
    return hierarchyName;
  }

  /**
   * Set the name of the hierarchy to use.
   * @param hierarchyName The name of the hierarchy to use.
   */
  public void setHierarchyName(String hierarchyName) {
    this.hierarchyName = hierarchyName;
  }

  /** The hierarchy manipulated by this Xml module. */
  protected UIH uih;

  /**
   * Get the hierarchy that was constructed by this XML module.
   * @return The hierarchy.
   */
  public UIH getUIH() {
    return uih;
  }

  /**
   * Construct a hierarchy of components.
   */
  public XmlUIH() {
    this(new UIH());
  }

  /**
   * Construct a hierarchy of components.
   * @param uih The hierarchy object to use to build the hierarchy from XML.
   */
  public XmlUIH(UIH uih) {
    this.uih = uih;
  }

  /**
   * Build a user interface containment hierarchy from an XML stream. It uses
   * the first hierarchy found in the XML stream.
   * @param xmlInputStream The stream containing the XML definitions.
   * @param parameters The optional parameters. They are in the form
   * {"name_of_component_1_in_xml", component_1,<br>
   *  "name_of_component_2_in_xml", component_2, ...}<br>
   * or:<br>
   * {component_1, component_2, ...} with the components' names set.<br>
   * or a mix of both.
   * @throws IOException If an IO error occurs.
   */
  public void build(InputStream xmlInputStream, Object[] parameters) throws IOException {
    build(xmlInputStream, new ArrayComponentResolver(parameters));
  }

  /**
   * Build a user interface containment hierarchy from an XML stream. It uses
   * the first hierarchy found in the XML stream.
   * @param xmlInputStream The stream containing the XML definitions.
   * @param componentResolver The component resolver used to get a component
   * from its name.
   * @throws IOException If an IO error occurs.
   */
  public void build(InputStream xmlInputStream, ComponentResolver componentResolver) throws IOException {
    setComponentResolver(componentResolver);
    buildHierarchy(xmlInputStream);
  }

  /**
   * Build a user interface containment hierarchy from an XML stream, using a
   * named hierachy.
   * @param xmlInputStream The stream containing the XML definitions.
   * @throws IOException If an IO error occurs.
   */
  protected void buildHierarchy(InputStream xmlInputStream) throws IOException {
    // Parse the XML document
    XMLElement element = new XMLElement(new Hashtable(), false, false);
    element.parseFromReader(new InputStreamReader(xmlInputStream));
    if(!element.getName().equals("uih")) {
      throw new IllegalStateException("One and only one \"uih\" element can be defined in the XML document!");
    }
    XMLElement hierarchyElement = null;
    String hierarchyName = getHierarchyName();
    for(Enumeration e = element.enumerateChildren(); e.hasMoreElements(); ) {
      XMLElement tmpElement = (XMLElement)e.nextElement();
      String name = tmpElement.getStringAttribute("name");
      if(hierarchyName == null || (name != null && name.equals(hierarchyName))) {
        hierarchyElement = tmpElement;
        break;
      }
    }
    if(hierarchyElement == null) {
      throw new IllegalStateException("The hierarchy to load is not defined in the document!");
    }
    for(Enumeration e = hierarchyElement.enumerateChildren(); e.hasMoreElements(); ) {
      element = (XMLElement)e.nextElement();
      createObjects(element);
    }
  }

  /**
   * Get the class object associated with a class name. It supports some
   * shortcuts for some known types.
   * @param className The name of the class to get.
   * @return The class corresponding to the class name.
   */
  protected Class getObjectClass(String className) {
    if(className == null) {
      return null;
    }
    if(Character.isLowerCase(className.charAt(0))) {
      if(className.equals("int")) {
        return int.class;
      } else if(className.equals("short")) {
        return short.class;
      } else if(className.equals("long")) {
        return long.class;
      } else if(className.equals("byte")) {
        return byte.class;
      } else if(className.equals("float")) {
        return float.class;
      } else if(className.equals("double")) {
        return double.class;
      } else if(className.equals("char")) {
        return char.class;
      } else if(className.equals("boolean")) {
        return boolean.class;
      }
    }
    if(className.indexOf(".") == -1) {
      if(className.endsWith("Layout")) {
        if(className.equals("BoxLayout") || className.equals("OverlayLayout") || className.equals("SpringLayout")) {
          className = "javax.swing." + className;
        } else if(className.equals("FormLayout")) {
          className = "com.jgoodies.forms.layout." + className;
        } else if(className.equals("ExplicitLayout")) {
          className = "com.zookitec.layout." + className;
        } else {
          className = "java.awt." + className;
        }
      } else {
        className = "java.lang." + className;
      }
    }
    try {
      return Class.forName(className);
    } catch(Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  /**
   * Get the object value when the class and String representation are known.
   * @param objectClass The class of the object.
   * @param value The string representation of the value.
   * @return The object value.
   */
  protected Object getObjectValue(Class objectClass, String value) {
    if(objectClass.equals(int.class)) {
      return new Integer(value);
    } else if(objectClass.equals(short.class)) {
      return new Short(value);
    } else if(objectClass.equals(long.class)) {
      return new Long(value);
    } else if(objectClass.equals(byte.class)) {
      return new Byte(value);
    } else if(objectClass.equals(float.class)) {
      return new Float(value);
    } else if(objectClass.equals(double.class)) {
      return new Double(value);
    } else if(objectClass.equals(char.class)) {
      return new Character(value.charAt(0));
    } else if(objectClass.equals(boolean.class)) {
      return new Boolean(value);
    }
    try {
      return objectClass.getConstructor(new Class[]{String.class}).newInstance(new String[] {value});
    } catch(Exception e) {
      return null;
    }
  }

  /**
   * Create the parameters from an element that contains parameters data.
   * @param parametersDataElement The XML element with the parameters.
   * @return The parameters, a 2-cell array containing an array of classes and
   * the array of values. Classes are return to differentiate cases like
   * "int.class" and "Integer.class".
   */
  protected Object[] createParameters(XMLElement parametersDataElement) {
    String className = parametersDataElement.getStringAttribute("class");
    String value = parametersDataElement.getStringAttribute("value");
    Object[] parameters = new Object[2];
    // Try to find if it is a single parameter
    if(className != null && value != null) {
      try {
        Class objectClass = getObjectClass(className);
        parameters[0] = new Class[] {objectClass};
        parameters[1] = new Object[] {objectClass.getConstructor(new Class[]{String.class}).newInstance(new String[] {value})};
      } catch(Exception e) {
        parameters[0] = new Class[] {};
        parameters[1] = new Object[] {};
      }
    } else {
      if(className != value) {
        throw new IllegalStateException("\"" + parametersDataElement.getName() + "\" must define both a class and a value!");
      }
      // Consider it is an array of attributes
      List dataList = new ArrayList();
      List classList = new ArrayList();
      for(Enumeration e = parametersDataElement.enumerateChildren(); e.hasMoreElements(); ) {
        XMLElement element = (XMLElement)e.nextElement();
        if(element.getName().equals("param")) {
          className = element.getStringAttribute("class");
          value = element.getStringAttribute("value");
          if(className != null && value != null) {
            try {
              Class objectClass = getObjectClass(className);
              classList.add(objectClass);
              dataList.add(getObjectValue(objectClass, value));
            } catch(Exception ex) {
              ex.printStackTrace();
            }
          }
        }
      }
      parameters[0] = new Class[classList.size()];
      parameters[1] = new Object[dataList.size()];
      for(int i=0; i<dataList.size(); i++) {
        ((Class[])parameters[0])[i] = (Class)classList.get(i);
        ((Object[])parameters[1])[i] = dataList.get(i);
      }
    }
    return parameters;
  }

  /**
   * Create necessary data for a component.
   * @param xmlElement The element from which to load necessary information.
   */
  protected void createObjects(XMLElement xmlElement) {
    // Get all the attributes of the component element
    String nodeType = xmlElement.getName();
    // Get the type of node
    boolean isRootNode = nodeType.equals("rootNode");
    boolean isNode = nodeType.equals("node");
    boolean isLeaf = nodeType.equals("leaf");
    if(!isRootNode && !isNode && !isLeaf) {
      throw new IllegalArgumentException("The element \"" + nodeType + "\" is not one of \"rootNode\", \"node\" or \"leaf\"!");
    }
    // Load the attributes
    String name = null;
    Object constraints = null;
    Class layoutClass = null;
    Object[] layoutParam = null;
    String value = null;
    String mappings = null;
    Object subConstraints = null;
    // Check that attributes are valid
    for(Enumeration e = xmlElement.enumerateAttributeNames(); e.hasMoreElements(); ) {
      String attributeName = (String)e.nextElement();
      if(attributeName.equals("name")) {
        name = xmlElement.getStringAttribute(attributeName);
      } else if(!isRootNode && attributeName.equals("constrain")) {
        constraints = xmlElement.getStringAttribute(attributeName);
      } else if(!isLeaf && attributeName.equals("layout")) {
        String layoutClassName = xmlElement.getStringAttribute(attributeName);
        if(layoutClassName != null) {
          layoutParam = new Object[2];
          String elementValue = xmlElement.getStringAttribute("layoutParam");
          layoutParam[0] = elementValue == null? null: new Class[] {String.class};
          layoutParam[1] = elementValue == null? null: new Object[] {elementValue};
          layoutClass = getObjectClass(layoutClassName);
        }
      } else if(isLeaf && attributeName.equals("value")) {
        value = xmlElement.getStringAttribute(attributeName);
      } else if(attributeName.equals("map")) {
        mappings = xmlElement.getStringAttribute(attributeName);
      } else if(!isLeaf && isRootNode && attributeName.equals("subConstrain")) {
        subConstraints = xmlElement.getStringAttribute(attributeName);
      } else {
        throw new IllegalStateException("The attribute \"" + attributeName + "\" is not supported by the element \"" + xmlElement.getName() + "\"!");
      }
    }
    // Additional check
    if(name != null && value != null) {
      throw new IllegalStateException("Only one of the name \"" + name + "\" or the value \"" + value + "\" can be specified!");
    }
    // The list of children
    List xmlElementsList = null;
    if(!isLeaf) {
      xmlElementsList = new ArrayList();
    }
    // Initialize the component using the parameters
    Object component = null;
    if(name != null) {
      component = componentResolver.resolve(name);
      if(component == null) {
        throw new IllegalArgumentException("The component \"" + name + "\" is not defined!");
      }
    }
    if(component == null) {
      component = value;
    }
    // Find the components and the sub components
    XMLElement layoutParamElement = null;
    XMLElement constraintsElement = null;
    XMLElement subConstraintsElement = null;
    for(Enumeration e = xmlElement.enumerateChildren(); e.hasMoreElements(); ) {
      XMLElement tmpElement = (XMLElement)e.nextElement();
      String eName = tmpElement.getName();
      if(xmlElementsList != null && (eName.equals("node") || eName.equals("leaf") || eName.equals("rootNode"))) {
        xmlElementsList.add(tmpElement);
      } else if(!isRootNode && eName.equals("constrain")) {
        if(constraints != null) {
          throw new IllegalStateException("The element \"" + xmlElement.getName() + "\" declares twice the constraints!");
        }
        constraintsElement = tmpElement;
      } else if(!isLeaf && eName.equals("layoutParam")) {
        if(layoutParamElement != null) {
          throw new IllegalStateException("The element \"" + xmlElement.getName() + "\" declares twice the layout parameters!");
        }
        layoutParamElement = tmpElement;
      } else if(!isLeaf && eName.equals("subConstrain")) {
        if(subConstraintsElement != null) {
          throw new IllegalStateException("The element \"" + xmlElement.getName() + "\" declares twice the sub constraints!");
        }
        subConstraintsElement = tmpElement;
      } else {
        throw new IllegalStateException("The contained element \"" + eName + "\" is not supported by the element \"" + xmlElement.getName() + "\"!");
      }
    }
    // Create the layout data
    if(layoutClass != null) {
      if(layoutParam[0] == null && layoutParamElement != null) {
        layoutParam = createParameters(layoutParamElement);
      }
      if(layoutParam[0] == null) {
        layoutParam[0] = new Class[] {};
        layoutParam[1] = new Object[] {};
      }
    }
    // Create the constraints
    if(constraintsElement != null) {
      if(constraints != null) {
        throw new IllegalStateException("The element \"" + xmlElement.getName() + "\" declares twice the constraints!");
      }
      constraints = ((Object[])createParameters(constraintsElement)[1])[0];
    }
    // Create the sub constraints
    if(subConstraintsElement != null) {
      if(subConstraints != null) {
        throw new IllegalStateException("The element \"" + xmlElement.getName() + "\" declares twice the sub constraints!");
      }
      subConstraints = ((Object[])createParameters(subConstraintsElement)[1])[0];
    }
    // Add to hierarchy
    if(isLeaf) {
      HLeaf leaf = uih.add(component);
      // constrain
      leaf.constrain(constraints);
      // map
      if(mappings != null) {
        leaf.map(mappings);
      }
    } else if(isRootNode) {
      HRootNode rootNode = uih.openNew(component);
      // layout
      LayoutManager layoutManager = null;
      if(layoutClass != null) {
        try {
          layoutManager = (LayoutManager)layoutClass.getConstructor((Class[])layoutParam[0]).newInstance((Object[])layoutParam[1]);
          rootNode.layout(layoutManager);
        } catch(Exception e) {
          throw new IllegalStateException("The layout manager for the element \"" + xmlElement.getName() + "\" could not be instantiated!");
        }
      }
      // subConstrain
      if(subConstraints != null) {
        rootNode.subConstrain(subConstraints);
      }
      // map
      if(mappings != null) {
        rootNode.map(mappings);
      }
    } else if(isNode) {
      HNode node = uih.open(component);
      // constrain
      node.constrain(constraints);
      // layout
      LayoutManager layoutManager = null;
      if(layoutClass != null) {
        try {
          layoutManager = (LayoutManager)layoutClass.getConstructor((Class[])layoutParam[0]).newInstance((Object[])layoutParam[1]);
          node.layout(layoutManager);
        } catch(Exception e) {
          throw new IllegalStateException("The layout manager for the element \"" + xmlElement.getName() + "\" could not be instantiated!");
        }
      }
      // subConstrain
      if(subConstraints != null) {
        node.subConstrain(subConstraints);
      }
      // map
      if(mappings != null) {
        node.map(mappings);
      }
    }
    // Do for children
    if(xmlElementsList != null && !xmlElementsList.isEmpty()) {
      for(int i=0; i<xmlElementsList.size(); i++) {
        createObjects((XMLElement)xmlElementsList.get(i));
      }
    }
    // Close the levels
    if(isRootNode || isNode) {
      uih.close();
    }
  }

}
