/*
 * @(#)FieldsComponentResolver.java
 * 
 * Christopher Deckers (chrriis@brainlex.com)
 * http://chrriis.brainlex.com
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package chrriis.uihierarchy.xml.resolver;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

import chrriis.uihierarchy.compatibility.Compatibility;
import chrriis.uihierarchy.compatibility.HashMap;
import chrriis.uihierarchy.compatibility.Map;

/**
 * The component resolver for object fields. It is able to resolve public
 * fields, but can also resolve other fields for JDK 1.2 and later if the
 * security manager allows changing member accessibility (via reflection).
 * @version 1.0 2003.10.06
 * @author Christopher Deckers (chrriis@brainlex.com)
 */
public class FieldsComponentResolver implements ComponentResolver {

  /** The object being considered. */
  protected Object object;
  /** The map of components. */
  protected Map componentsMap = new HashMap();

  /**
   * Construct a fields component resolver for an object.
   * @param object The object from which the fields will be inspected and added
   * to the resolve mechanism.
   */
  public FieldsComponentResolver(Object object) {
    this.object = object;
  }

  /**
   * Resolve a component from its name.
   * @param name The name of the component to resolve.
   * @see chrriis.uihierarchy.xml.resolver.ComponentResolver#resolve(java.lang.String)
   */
  public Object resolve(String name) {
    Object component = componentsMap.get(name);
    if(component != null) {
      return component;
    }
    // Subsequent calls of a non existing field will return null directly.
    if(componentsMap.containsKey(name)) {
      return null;
    }
    try {
      Field field = object.getClass().getDeclaredField(name);
      int modifiers = field.getModifiers();
      if(!Compatibility.isJava1_1() && !Modifier.isPublic(modifiers)) {
        field.setAccessible(true);
      }
      component = field.get(object);
      if(!Compatibility.isJava1_1() && !Modifier.isPublic(modifiers)) {
        field.setAccessible(false);
      }
    } catch(Exception e) {}
    componentsMap.put(name, component);
    return component;
  }

}
